using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using System.IO;

namespace WindowsGame1
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        // Gal: I applogize in advance for the lack 
        // pretty much writing the entire game in one class

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        Vector2 mPosition = new Vector2(0, 0);
        Vector2 mImagePosition;
        Texture2D mSpriteTexture;
        Texture2D mSpriteFillTexture;
        Vector2 mSpeed = Vector2.Zero;
        Texture2D mImageTexture;
        Texture2D mMaskTexture;
        Random mRandom = new Random();
        Color mPlayerCurrentColor;
        Vector2 mScreenCenter;
        Vector2 mImageCenter;
        ColorCircle[] mColorCircles;
        TimeSpan? mTimeToRandomizeColors;
        HashSet<Point> mEnemyPositions;
        Color[] mMaskPixels;
        SpriteFont mFont;
        int mScore = 0;
        float mPercentScore = 0;
        float mBestPercentScore = 0;
        TimeSpan mTimeSinceLastScoreUpdate = TimeSpan.Zero;
        bool mIsInSuperMode;
        

        const int PLAYER_RING_RADIUS = 10;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 1000;
            graphics.PreferredBackBufferHeight = 1000;
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            mPlayerCurrentColor = Color.FromNonPremultiplied(255,255,255,0);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // TODO: use this.Content to load your game content here
            mFont = Content.Load<SpriteFont>("SpriteFont1");

            mScreenCenter = new Vector2(
                GraphicsDevice.Viewport.Bounds.Width / 2,
                GraphicsDevice.Viewport.Bounds.Height / 2);

            mSpriteTexture = CreateCircleStroke(PLAYER_RING_RADIUS);

            //mImageTexture = this.Content.Load<Texture2D>("1596807_10151986495878311_2059955370_o - Copy");
            //mImageTexture = Texture2D.FromStream(GraphicsDevice, File.OpenRead("Image800_600.jpg"));
            try
            {
                mImageTexture = Crop(Texture2D.FromStream(GraphicsDevice, File.OpenRead("Image.jpg")), new Rectangle(0, 0, 800, 600));
            }
            catch
            {
                mImageTexture = Crop(this.Content.Load<Texture2D>("Image"), new Rectangle(0, 0, 800, 600));
            }
            mMaskTexture = new Texture2D(GraphicsDevice, mImageTexture.Bounds.Width, mImageTexture.Bounds.Height);

            mEnemyPositions = new HashSet<Point>();
            
            int[] pixels = new int[mMaskTexture.Width * mMaskTexture.Height];
            for (int t = 0; t < pixels.Length - 1; t++)
            {
                pixels[t] = unchecked((int)0xffffffff);
            }
            mMaskTexture.SetData(pixels);

            mMaskPixels = new Color[mMaskTexture.Width * mMaskTexture.Height];
            mMaskTexture.GetData<Color>(mMaskPixels);
            
            mImageCenter = new Vector2(
                mImageTexture.Width / 2,
                mImageTexture.Height / 2);

            mImagePosition = mScreenCenter - mImageCenter;
            mPosition = mScreenCenter;

            mColorCircles = new ColorCircle[4];
            mColorCircles[0] = new ColorCircle() { Position = new Vector2(mImagePosition.X - 10 - 20, mScreenCenter.Y), CircleColor = GetRandomColor(), Texture = CreateCircleFill(10) };
            mColorCircles[1] = new ColorCircle() { Position = new Vector2(mImagePosition.X + mImageTexture.Width + 10, mScreenCenter.Y), CircleColor = GetRandomColor(), Texture = CreateCircleFill(10) };
            mColorCircles[2] = new ColorCircle() { Position = new Vector2(mScreenCenter.X, mImagePosition.Y - 10 - 20), CircleColor = GetRandomColor(), Texture = CreateCircleFill(10) };
            mColorCircles[3] = new ColorCircle() { Position = new Vector2(mScreenCenter.X, mImagePosition.Y + mImageTexture.Height + 10), CircleColor = GetRandomColor(), Texture = CreateCircleFill(10) };
        }

        public static Texture2D Crop(Texture2D image, Rectangle source)
        {
            var graphics = image.GraphicsDevice;
            var ret = new RenderTarget2D(graphics, source.Width, source.Height);
            var sb = new SpriteBatch(graphics);

            graphics.SetRenderTarget(ret); // draw to image
            graphics.Clear(new Color(0, 0, 0, 0));

            sb.Begin();
            sb.Draw(image, Vector2.Zero, source, Color.White);
            sb.End();

            graphics.SetRenderTarget(null); // set back to main window

            return (Texture2D)ret;
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here
            UpdateMovement(Keyboard.GetState());
            mPosition += mSpeed * (float)gameTime.ElapsedGameTime.TotalSeconds;
            mPosition.X = Math.Max(Math.Min(mPosition.X, GraphicsDevice.Viewport.Bounds.Width), 0);
            mPosition.Y = Math.Max(Math.Min(mPosition.Y, GraphicsDevice.Viewport.Bounds.Height), 0);

            mSpriteFillTexture = CreateCircleFill(PLAYER_RING_RADIUS /2);
            
            mIsInSuperMode = Keyboard.GetState().IsKeyDown(Keys.Space);
            CheckPlayerHitColorCircles();
            
            HideMaskAroundPlayer();

            if (mTimeToRandomizeColors != null)
            {
                mTimeToRandomizeColors = mTimeToRandomizeColors.Value.Subtract(gameTime.ElapsedGameTime);
                if (mTimeToRandomizeColors.Value < TimeSpan.Zero)
                {
                    mTimeToRandomizeColors = null;
                    RandomizeColorCircles();
                }
            }

            UpdateEnemy();

            mTimeSinceLastScoreUpdate = mTimeSinceLastScoreUpdate.Add(gameTime.ElapsedGameTime);
            if (mTimeSinceLastScoreUpdate > TimeSpan.FromSeconds(0.5))
            {
                UpdateScore();
                mTimeSinceLastScoreUpdate = TimeSpan.Zero;
            }

            mMaskTexture.SetData(mMaskPixels);

            base.Update(gameTime);
        }

        private void UpdateScore()
        {
            int score = 0;
            foreach (Color color in mMaskPixels)
            {
                if (color.A <= 0)
                {
                    score++;
                }
            }
            mScore = score;
            mPercentScore = (float)score / mMaskPixels.Length;
            mBestPercentScore = Math.Max(mBestPercentScore, mPercentScore);
        }

        private void UpdateEnemy()
        {            
            Point randPoint = new Point(mRandom.Next(mImageTexture.Bounds.Width), mRandom.Next(mImageTexture.Bounds.Height));

            if (mRandom.NextDouble() > 0.995 - mPercentScore / 50)
            {
                TryCreateEnemy(randPoint);
            }

            if (mEnemyPositions.Count == 0) { return; }

            List<Point> currentEnemyPositions = mEnemyPositions.ToList();

            Point point;
            for (int i = 0; i < currentEnemyPositions.Count / 2; i++)
            {
                if (mRandom.NextDouble() < 0.99) { continue; }

                point = currentEnemyPositions[i];

                Point expansionPoint = new Point(point.X + (mRandom.Next(3) - 1) * 10, point.Y + (mRandom.Next(3) - 1) * 10);

                if (currentEnemyPositions.Contains(expansionPoint))
                {
                    currentEnemyPositions.Remove(point);
                    currentEnemyPositions.Remove(expansionPoint);
                    if (currentEnemyPositions.Count == 0) { break; }

                    i--;
                    continue;
                }
                TryCreateEnemy(expansionPoint);
            }
        }

        private void TryCreateEnemy(Point aPoint)
        {
            int pixelIndex = aPoint.Y * mMaskTexture.Width + aPoint.X;

            if (pixelIndex < 0 || pixelIndex > mMaskTexture.Width * mMaskTexture.Height - 1) { return; }

            if (mMaskPixels[pixelIndex].A > 0)
            {
                mEnemyPositions.Add(aPoint);

                int radius = 8;
                float angleStep = 0.1f;
                int outerRadius = radius * 2 + 2;
                for (float rad = radius; rad > 0; rad -= 0.5f)
                {
                    for (double angle = 0; angle < Math.PI * 2; angle += angleStep)
                    {
                        int x = aPoint.X + (int)Math.Round(radius + rad * Math.Cos(angle));
                        int y = aPoint.Y + (int)Math.Round(radius + rad * Math.Sin(angle));

                        int pixelIndex2 = y * mMaskTexture.Width + x + 1;

                        if (pixelIndex2 < 0 || pixelIndex2 > mMaskTexture.Width * mMaskTexture.Height - 1) { continue; }

                        mMaskPixels[pixelIndex2] = Color.Black;
                    }
                }
            }
            else
            {
                mMaskPixels[pixelIndex].A++;
            }
        }

        private void CheckPlayerHitColorCircles()
        {
            if (mTimeToRandomizeColors != null) { return; }

            Rectangle playerBounds = mSpriteTexture.Bounds;
            playerBounds.Offset((int)mPosition.X, (int)mPosition.Y);

            for (int index = 0; index < mColorCircles.Length; index++)
            {
                Rectangle circleBounds = mColorCircles[index].Texture.Bounds;
                circleBounds.Offset((int)mColorCircles[index].Position.X, (int)mColorCircles[index].Position.Y);

                if (circleBounds.Intersects(playerBounds))
                {
                    mPlayerCurrentColor = mColorCircles[index].CircleColor;
                    HideColorCircles();
                    mTimeToRandomizeColors = TimeSpan.FromSeconds(2);
                    break;
                }
            }
        }

        private void UpdateMovement(KeyboardState keyboardState)
        {
            const float ACCELERATION = 50f;
            const float DAMPING_FACTOR = 0.9f;
            const float MAX_SPEED = 400f;

            if (keyboardState.IsKeyDown(Keys.Left) == true)
            {
                mSpeed.X -= ACCELERATION;
            }

            else if (keyboardState.IsKeyDown(Keys.Right) == true)
            {
                mSpeed.X += ACCELERATION;
            }

            if (keyboardState.IsKeyDown(Keys.Up) == true)
            {
                mSpeed.Y -= ACCELERATION;
            }

            else if (keyboardState.IsKeyDown(Keys.Down) == true)
            {
                mSpeed.Y += ACCELERATION;
            }

            mSpeed.X *= DAMPING_FACTOR;
            mSpeed.Y *= DAMPING_FACTOR;

            mSpeed.X = Math.Max(Math.Min(mSpeed.X, MAX_SPEED), -MAX_SPEED);
            mSpeed.Y = Math.Max(Math.Min(mSpeed.Y, MAX_SPEED), -MAX_SPEED);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.White);
            
            // TODO: Add your drawing code here
            spriteBatch.Begin();
            spriteBatch.Draw(mImageTexture, mImagePosition, mImageTexture.Bounds, Color.White);
            spriteBatch.Draw(mMaskTexture, mImagePosition, mImageTexture.Bounds, Color.White);
            for (int colorCircleIndex = 0; colorCircleIndex < mColorCircles.Length; colorCircleIndex++) { mColorCircles[colorCircleIndex].Draw(spriteBatch); }
            spriteBatch.Draw(mSpriteTexture, mPosition - new Vector2(PLAYER_RING_RADIUS, PLAYER_RING_RADIUS), Color.Black);
            spriteBatch.Draw(mSpriteFillTexture, mPosition - new Vector2(PLAYER_RING_RADIUS/2, PLAYER_RING_RADIUS/2), mPlayerCurrentColor);
            spriteBatch.DrawString(mFont, string.Format("Score: {0:p} | Best: {1:p} | Creep: {2}",mPercentScore,mBestPercentScore, mEnemyPositions.Count), mImagePosition - new Vector2(0, 50), Color.Black);

            spriteBatch.End();

            base.Draw(gameTime);
        }

        private void HideMaskAroundPlayer()
        {
            double r;
            double ang;
            Point pointPosition;

            Color[] imagePixels = new Color[mMaskTexture.Width * mMaskTexture.Height];
            mImageTexture.GetData<Color>(imagePixels);

            Vector2 positionOnImage = mPosition - mImagePosition;

            for (int t = 0; t < 1000; t++)
            {
                r = mRandom.NextDouble() * 100 + PLAYER_RING_RADIUS;

                if (mIsInSuperMode) { r /= 4; }

                ang = mRandom.NextDouble() * Math.PI * 2;

                pointPosition = new Point((int)(positionOnImage.X + Math.Cos(ang) * r), (int)(positionOnImage.Y + Math.Sin(ang) * r));

                if (pointPosition.X < 0 || pointPosition.X > mImageTexture.Bounds.Width || pointPosition.Y < 0 || pointPosition.Y > mImageTexture.Bounds.Height) { continue; }

                //int pixelIndex = pointPosition.Y * mMaskTexture.Width + pointPosition.X;
                int pixelIndex;
                if (!PointToIndex(pointPosition, out pixelIndex)) { continue; }

                int distance = GetColorsDistance(imagePixels[pixelIndex], mPlayerCurrentColor);
                if (distance < 150 || mIsInSuperMode)
                {
                    if (!mIsInSuperMode)
                    {
                        mMaskPixels[pixelIndex] = Color.FromNonPremultiplied(255, 255, 255, Math.Max(mMaskPixels[pixelIndex].A - (150 - distance), 0));

                        ChangeColorIfValid(OffsetPoint(pointPosition, 0, -1), mMaskPixels[pixelIndex]);
                        ChangeColorIfValid(OffsetPoint(pointPosition, 0, 1), mMaskPixels[pixelIndex]);
                        ChangeColorIfValid(OffsetPoint(pointPosition, -1, 0), mMaskPixels[pixelIndex]);
                        ChangeColorIfValid(OffsetPoint(pointPosition, 1, 0), mMaskPixels[pixelIndex]);
                    }
                    else
                    {
                        mMaskPixels[pixelIndex] = Color.FromNonPremultiplied(255, 255, 255, 255);
                    }

                    if (mEnemyPositions.Contains(pointPosition)) { mEnemyPositions.Remove(pointPosition); }
                }
            }
        }

        private bool PointToIndex(Point aPoint, out int aIndex)
        {
            aIndex = aPoint.Y * mImageTexture.Width + aPoint.X;
            if (aIndex < 0 || aIndex > mImageTexture.Width * mImageTexture.Height - 1)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        private Point OffsetPoint(Point aPoint, int aX, int aY)
        {
            return new Point(aPoint.X + aX, aPoint.Y + aY);
        }

        private void ChangeColorIfValid(Point aPoint, Color aColor)
        {
            int index;
            bool valid = PointToIndex(aPoint, out index);
            if (valid)
            {
                mMaskPixels[index] = aColor;
            }
        }

        private int GetColorsDistance(Color aColorA, Color aColorB)
        {
            return Math.Abs(aColorA.R - aColorB.R) +
                   Math.Abs(aColorA.G - aColorB.G) +
                   Math.Abs(aColorA.B - aColorB.B) +
                   Math.Abs(aColorA.A - aColorB.A);
        }

        public Texture2D CreateCircleStroke(int radius)
        {
            int outerRadius = radius * 2 + 2; // So circle doesn't go out of bounds
            Texture2D texture = new Texture2D(GraphicsDevice, outerRadius, outerRadius);

            Color[] data = new Color[outerRadius * outerRadius];

            // Colour the entire texture transparent first.
            for (int i = 0; i < data.Length; i++)
                data[i] = Color.FromNonPremultiplied(255,255,255,0);

            // Work out the minimum step necessary using trigonometry + sine approximation.
            double angleStep = 1f / radius;

            for (double angle = 0; angle < Math.PI * 2; angle += angleStep)
            {
                // Use the parametric definition of a circle: http://en.wikipedia.org/wiki/Circle#Cartesian_coordinates
                int x = (int)Math.Round(radius + radius * Math.Cos(angle));
                int y = (int)Math.Round(radius + radius * Math.Sin(angle));

                data[y * outerRadius + x + 1] = Color.White;
            }

            texture.SetData(data);
            return texture;
        }

        public Texture2D CreateCircleFill(int radius)
        {
            int outerRadius = radius * 2 + 2; // So circle doesn't go out of bounds
            Texture2D texture = new Texture2D(GraphicsDevice, outerRadius, outerRadius);

            Color[] data = new Color[outerRadius * outerRadius];

            // Colour the entire texture transparent first.
            for (int i = 0; i < data.Length; i++)
                data[i] = Color.FromNonPremultiplied(255, 255, 255, 0);

            // Work out the minimum step necessary using trigonometry + sine approximation.
            double angleStep = 1f / radius;

            for (float rad = radius; rad > 0; rad-=0.5f)
            {
                for (double angle = 0; angle < Math.PI * 2; angle += angleStep)
                {
                    int x = (int)Math.Round(radius + rad * Math.Cos(angle));
                    int y = (int)Math.Round(radius + rad * Math.Sin(angle));

                    data[y * outerRadius + x + 1] = Color.White;
                }
            }

            texture.SetData(data);
            return texture;
        }

        private Color GetRandomColor()
        {
            return Color.FromNonPremultiplied((byte)mRandom.Next(255), (byte)mRandom.Next(255), (byte)mRandom.Next(255), 255);
        }

        private void RandomizeColorCircles()
        {
            for (int colorCircleIndex = 0; colorCircleIndex < mColorCircles.Length; colorCircleIndex++)
            {
                mColorCircles[colorCircleIndex].CircleColor = GetRandomColor();
            }
        }

        private void HideColorCircles()
        {
            for (int colorCircleIndex = 0; colorCircleIndex < mColorCircles.Length; colorCircleIndex++)
            {
                mColorCircles[colorCircleIndex].CircleColor = Color.FromNonPremultiplied(255, 255, 255, 0);
            }
        }
    }
}
